@@ -67,10 +67,12 @@ const std::string deviceDevtype   = "DEVTYPE=thunderbolt_device";
 // TODO: replace with boost.program_options
 const std::string opt_devices     = "devices";
 const std::string opt_topology    = "topology";
+const std::string opt_approve     = "approve";
 const std::string opt_approve_all = "approve-all";
 const std::string opt_acl         = "acl";
 const std::string opt_remove      = "remove";
 const std::string opt_remove_all  = "remove-all";
+const std::string opt_once_flag   = "--once";
 
 const std::string indent     = "|   ";
 const std::string indentLast = "    ";
@@ -87,6 +89,39 @@ std::string readAndTrim(const fs::path& path)
     return boost::algorithm::trim_copy(read(path));
 }
 
+/**
+ * Return the content of the file from given path or "Unknown" + type if empty
+ *
+ * @param path  path to file to read from
+ * @param type  string to return if file is empty, prefixed with "Unknown"
+ */
+std::string readName(const fs::path& path, const std::string& type)
+{
+    try
+    {
+        auto res = readAndTrim(path);
+        if (!res.empty())
+        {
+            return res;
+        }
+    }
+    catch (std::runtime_error&)
+    {
+        // assuming this is from an empty file
+    }
+    return "Unknown " + type;
+}
+
+std::string readVendor(const fs::path& path)
+{
+    return readName(path, "vendor");
+}
+
+std::string readDevice(const fs::path& path)
+{
+    return readName(path, "device");
+}
+
 bool findUeventAttr(const fs::path& path, const std::string& attribute)
 {
     const auto ueventFile = path / "uevent";
@@ -182,8 +217,24 @@ void tbtadm::Controller::run()
         {
             return topology();
         }
+        if (m_argv[1] == opt_approve)
+        {
+            if (m_argc == 3 || m_argc == 4)
+            {
+                if (m_argv[2] == opt_once_flag)
+                {
+                    m_once = true;
+                }
+                m_sl = findSL();
+                return approve(sysfsDevicesPath / m_argv[m_argc - 1]);
+            }
+        }
         if (m_argv[1] == opt_approve_all)
         {
+            if (m_argc == 3 && m_argv[2] == opt_once_flag)
+            {
+                m_once = true;
+            }
             return approveAll();
         }
         if (m_argv[1] == opt_acl)
@@ -206,8 +257,10 @@ void tbtadm::Controller::run()
     // TODO: help
     const std::string sep = " | ";
     m_out << "Usage: " << opt_devices << sep << opt_topology << sep
-          << opt_approve_all << sep << opt_acl << sep << opt_remove
-          << " <uuid>|<route-string>" << sep << opt_remove_all << "\n";
+          << opt_approve << '[' << opt_once_flag << ']' << sep
+          << opt_approve_all << '[' << opt_once_flag << ']' << sep << opt_acl
+          << sep << opt_remove << " <uuid>|<route-string>" << sep
+          << opt_remove_all << "\n";
     throw std::runtime_error("Wrong usage");
 }
 
@@ -249,8 +302,8 @@ void tbtadm::Controller::devices()
 
         // TODO: better formatting
         const auto routeString = dir.path().filename().string();
-        m_out << routeString << '\t' << readAndTrim(vendorFilename) << '\t'
-              << readAndTrim(deviceFilename) << '\t' << authorized() << '\t'
+        m_out << routeString << '\t' << readVendor(vendorFilename) << '\t'
+              << readDevice(deviceFilename) << '\t' << authorized() << '\t'
               << inACL() << '\n';
     }
 }
@@ -284,8 +337,9 @@ void tbtadm::Controller::topology()
             m_sl          = slMap.find(readAndTrim(security))->second.num;
             std::vector<std::string> desc;
             desc.emplace_back("Controller "s + num + '\n');
-            desc.emplace_back("Name: " + readAndTrim(p / deviceFilename) + ", "
-                              + read(p / vendorFilename));
+            desc.emplace_back("Name: " + readDevice(p / deviceFilename) + ", "
+                              + readVendor(p / vendorFilename)
+                              + '\n');
             desc.emplace_back("Security level: "
                               + slMap.find(readAndTrim(security))->second.desc
                               + '\n');
@@ -341,8 +395,9 @@ void tbtadm::Controller::createTree(ControllerInTree& controller,
 
         auto routeString = p.filename().string();
         std::vector<std::string> desc;
-        desc.emplace_back(readAndTrim(p / deviceFilename) + ", "
-                          + read(p / vendorFilename));
+        desc.emplace_back(readDevice(p / deviceFilename) + ", "
+                          + readVendor(p / vendorFilename)
+                          + "\n");
         desc.emplace_back("Route-string: " + routeString + "\n");
         desc.emplace_back("Authorized: " + authorized(p) + "\n");
         desc.emplace_back("In ACL: " + inACL(p) + "\n");
@@ -441,7 +496,10 @@ void tbtadm::Controller::approve(const fs::path& dir) try
         return;
     }
 
-    addToACL();
+    if (!m_once)
+    {
+        addToACL();
+    }
 
     std::ostringstream keyStream;
     if (m_sl == 2)
@@ -461,7 +519,7 @@ void tbtadm::Controller::approve(const fs::path& dir) try
     authorized << 1;
 
     m_out << "Authorized\n";
-    if (m_sl == 2)
+    if (m_sl == 2 && !m_once)
     {
         File keyACL(acltree / readAndTrim(uniqueIDFilename) / keyFilename,
                     File::Mode::Write,
@@ -543,8 +601,8 @@ void tbtadm::Controller::acl()
         if (m_sl != 2 || fs::exists(p / keyFilename))
         {
             const auto uuid = p.filename().string();
-            m_out << uuid << '\t' << readAndTrim(p / vendorFilename) << '\t'
-                  << readAndTrim(p / deviceFilename) << '\t' << connected(uuid)
+            m_out << uuid << '\t' << readVendor(p / vendorFilename) << '\t'
+                  << readDevice(p / deviceFilename) << '\t' << connected(uuid)
                   << '\n';
         }
         else
@@ -562,8 +620,8 @@ void tbtadm::Controller::acl()
             if (!fs::exists(p / keyFilename))
             {
                 const auto uuid = p.filename().string();
-                m_out << uuid << '\t' << readAndTrim(p / vendorFilename) << '\t'
-                      << readAndTrim(p / deviceFilename) << '\t'
+                m_out << uuid << '\t' << readVendor(p / vendorFilename) << '\t'
+                      << readDevice(p / deviceFilename) << '\t'
                       << connected(uuid) << '\n';
             }
         }
